/***********************************************************************
**  Device driver for Oregon Micro System's PC-58
**  An Intelligent Motor Controller.
**  Linux 2.0 Operating System.
** --------------------------------------------------------------------
**  Written by:
**     Tony Denault
**     Institute for Astronomy, University of Hawaii
**     2680 Woodlawn Drive, Honolulu, HI 96822
**     Ph: (808)956-8101
**     email: denault@irtf.ifa.hawaii.edu
************************************************************************
*/

/*------------------------------
** include files
**------------------------------
*/

#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/malloc.h>

#include <linux/pc58.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

/*------------------------------------
** device defines
**-----------------------------------
*/
#define PC58_MAJOR_NO  26   /* major device number */
#define PC58_NO         1   /* number of minor devices */

#define PC58_VERSION  0010  /* This driver's version number 0.10  */
#define DEBUG 0

/*----------------------------------------
**  Structures
**----------------------------------------
*/

/*-----------------------------
** prototype functions
**-----------------------------
*/
static int pc58_configure ( int minor );

static int  pc58_open( struct inode * inode, struct file * file);
static void pc58_release( struct inode * inode, struct file * file);
static int  pc58_ioctl( struct inode *inode, struct file *file,
   unsigned int cmd, unsigned long arg);

/* low level functions */
static void PC58csleep( int csec );
static int PC58_puts( int minor, char *s );
static int PC58_gets( int minor, char *s, int s_len );
static int PC58_clear_inbuf( int minor );
static int PC58_command( int minor, char *command, char *reply,  int r_len );
static int PC58_read_axis_info( int minor, int axis_inx, struct pc58_dram_axis_t *dest );

/*-----------------------------
** global variable
**-----------------------------
*/

static struct pc58_table_t pc58_table[PC58_NO] =
{
   {
      0x310,   /* base            */
		0xd8000, /* shared mem address */
		0,       /* slip  reg */
		0,       /* done  reg */
		0,       /* limit reg */
		0,       /* home  reg */
   },
};

/* Define device driver entries points */
struct file_operations pc58_fops =
{
   NULL,        /* _seek    */
   NULL,        /* _read    */
   NULL,        /* _write   */
   NULL,        /* _readdir */
   NULL,        /* _select  */
   pc58_ioctl,  /* _ioctl   */
   NULL,        /* _mmap    */
   pc58_open,    /* _open    */
   pc58_release, /* _release */
};

static struct semaphore Sem = MUTEX;

/*-------------------------------------------------------------
** pc58_init() - Called at time the device driver is loaded
**-------------------------------------------------------------
*/
int pc58_init ( void )
{
   int err, 
	    minor;

	printk("Oregon Micro System PC-58 Motor Controller: major %d\n", PC58_MAJOR_NO);

   /* register device & request a major number */

   err = register_chrdev( PC58_MAJOR_NO, "pc58", &pc58_fops);
   if( err < 0 )
   {
		if(DEBUG) printk( KERN_NOTICE "pc58_init(): error %d\n", -err);
      return err;
   }

   /* Configure the device */
   for( minor=0; minor < PC58_NO; minor++ )
   {
		/* identify the IOport used by the device  */
		request_region( pc58_table[minor].base, PC58_NUM_IOPORTS, "pc58");

	   if( (err = pc58_configure( minor )) < 0 )
	      printk("   PC-58 minor %d - Initialization ERROR\n", minor );
		else
	      printk("   PC-58 minor %d - OK! base=0x%x shared_mem=0x%x\n",
			   minor, pc58_table[minor].base, pc58_table[minor].shared_mem);
   }
	return 0;
}

/*-------------------------------------------------------------
** pc58_configure()  each minor device
**-------------------------------------------------------------
*/
int pc58_configure ( int minor )
{
   int cnt, ready;

	/*-------------------------------------------------
	** wait until the board's intializated flag is set 
	*/
	cnt   = 0;
	ready = inb_p( pc58_table[minor].base+PC58_STATUS_REG) & PC58_STATUS_REG_initialized;
	while ( !ready )
	{
	   if( ++cnt > 10 )
		   return -1;
		PC58csleep( 1 );
	   ready = inb_p( pc58_table[minor].base+PC58_STATUS_REG) & PC58_STATUS_REG_initialized;
	}
	if(DEBUG) printk( "pc58_config(): cnt = %d \n", cnt);

	/*-------------------------------------------------
	** Configure PC58 - interrupts & shared memory address
	*/
	outb_p( 0x00, pc58_table[minor].base + PC58_CNTL_REG);

   /* set share memory address */
	outb_p( ((pc58_table[minor].shared_mem >> 16) & 0xff),
	        pc58_table[minor].base + PC58_DPRAM_HI_REG);

	outb_p( ((pc58_table[minor].shared_mem >> 8) & 0xff),
	        pc58_table[minor].base + PC58_DPRAM_LO_REG);

   /*---------------------------------------------------
	** set shared memory variable related to io buffer.
	** sometime they come up funny. 
	*/
	{
		struct pc58_dram_t * sm = (struct pc58_dram_t *) pc58_table[minor].shared_mem;  
		sm->out_write_inx =  sm->out_inx_end; 
		sm->in_read_inx = sm->in_inx_end; 
	}

   return 0;
}

/*-------------------------------------------------------------
**  Driver _open() routine. Called when process open driver.
**-------------------------------------------------------------
*/
static int pc58_open( struct inode * inode, struct file * file)
{
   if(DEBUG) printk("pc58_open(): enter\n");
   return 0;
}

/*-------------------------------------------------------------
**  Driver _release() routine. Called when process closes driver.
**-------------------------------------------------------------
*/
static void pc58_release( struct inode * inode, struct file * file)
{
   if(DEBUG) printk("pc58_release(): enter\n");
}

/*-------------------------------------------------------------
**  Driver _ioctl() routine.
**-------------------------------------------------------------
*/
static int pc58_ioctl( struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg)
{
   struct pc58parms_t pc58;   /*  driver's copy of user's parameters */
   int minor;

   minor = MINOR(inode->i_rdev);
   if( minor >= PC58_NO )
      return -ENODEV;

   /*--------------------------------------------------
   ** Read status register & OR bits into config table
   */
	pc58_table[minor].status_reg = inb_p( pc58_table[minor].base + PC58_STATUS_REG );
	pc58_table[minor].slip_reg   |= inb_p( pc58_table[minor].base + PC58_SLIP_REG );
	pc58_table[minor].done_reg   |= inb_p( pc58_table[minor].base + PC58_DONE_REG );
	pc58_table[minor].limit_reg  |= inb_p( pc58_table[minor].base + PC58_LIMIT_REG );
	pc58_table[minor].home_reg   |= inb_p( pc58_table[minor].base + PC58_HOME_REG );

	pc58_table[minor].dio_lo    = inb_p( pc58_table[minor].base + PC58_DIO_LO_REG );
	pc58_table[minor].dio_hi    = inb_p( pc58_table[minor].base + PC58_DIO_HI_REG );

   /*---------------------------------------
   ** copy pc58parms_t into kernel variable
   */
   memcpy_fromfs( &pc58, (struct pc58parms_t *)arg, sizeof(struct pc58parms_t));

   switch ( cmd )
   {
      case PC58Test:
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58GetConfig:
         pc58.ret = PC58_ERR_NONE;
         memcpy_tofs( pc58.p[0], &pc58_table[minor], sizeof(struct pc58_table_t));
         break;

      case PC58MutTest:
		   down( &Sem);
         PC58csleep( (int)pc58.l[0] );
         up( &Sem );
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58Inb:
         pc58.l[1] = inb_p(pc58.l[0]);
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58Outb:
         outb_p(pc58.l[1], pc58.l[0]);
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58GetSMValues:
		   {
			   struct pc58_dram_t * sm = (struct pc58_dram_t *) pc58_table[minor].shared_mem;  
            pc58.l[0] = sm->in_inx_end; 
            pc58.l[1] = sm->out_inx_end; 

            pc58.l[2] = sm->out_write_inx; 
            pc58.l[3] = sm->in_read_inx; 
			}
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58Puts:
         down( &Sem );
		   pc58.ret = PC58_puts( minor, pc58.p[0]);
         up( &Sem );
         break;

      case PC58Gets:
         down( &Sem );
		   pc58.ret = PC58_gets( minor, pc58.p[0], pc58.l[0]);
         up( &Sem );
         break;

      case PC58Command:
         down( &Sem );
		   pc58.ret = PC58_command( minor, pc58.p[0], pc58.p[1], pc58.l[0]);
         up( &Sem );
         break;

      case PC58ReadAxisInfo:
         down( &Sem );
			pc58.ret = PC58_read_axis_info( minor, pc58.l[0], pc58.p[0] );
         up( &Sem );
         break;

      case PC58ClearSlipFlag:
	      pc58_table[minor].slip_reg &= ~(0xff & pc58.l[0]);
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58ClearDoneFlag:
	      pc58_table[minor].done_reg &= ~(0xff & pc58.l[0]);
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58ClearLimitFlag:
	      pc58_table[minor].limit_reg &= ~(0xff & pc58.l[0]);
         pc58.ret = PC58_ERR_NONE;
         break;

      case PC58ClearHomeFlag:
	      pc58_table[minor].home_reg &= ~(0xff & pc58.l[0]);
         pc58.ret = PC58_ERR_NONE;
         break; 

       default:
         pc58.ret = PC58_ERR_INV_FUN;
         break;
   }

   /*---------------------------------------
   ** copy pc58parms_t into user space
   */
   memcpy_tofs( (struct pc58parms_t *)arg, &pc58, sizeof(struct pc58parms_t));
   return 0;
}

/*************************************************************************
*   Static function for device driver 
**************************************************************************/

/*--------------------------------------------------------------
**  PC58csleep() - a sleep function in units fo 1/100 seconds.
**                csec = 1 means sleep for 0.01 seconds.
**--------------------------------------------------------------
*/
void PC58csleep( int csec )
{
   current->state = TASK_INTERRUPTIBLE;
   current->timeout = jiffies + csec;
   schedule();
}

/*-------------------------------------------------------------------------
**  PC58_puts() - writes a string to the PC-58 using shared mem interface.
**-------------------------------------------------------------------------
*/
int PC58_puts( int minor, char *s )
{
   int inx, cnt, ch;

   struct pc58_dram_t *sm = (struct pc58_dram_t *) pc58_table[minor].shared_mem;

   inx = sm->out_write_inx;  /* get start of output buffer */

	while( (ch = get_user(s++)) )
	{
	   sm->outbuf[inx] = ch;    /* write char to buffer */
		
		if( ++inx >= PC58_BUF_SIZE) /* increment index to next location */
		   inx = 0;
	   
		/* if the buffer is full, just wait (but not too long) */
		cnt = 0;
		while( inx == sm->out_inx_end )
		{
		   PC58csleep( 1 );
		   if( ++cnt > 5 )
			   return PC58_ERR_TIMEOUT;
		}

      /* update out_write_inx (notifies PC58 a char is ready) */
		sm->out_write_inx = inx;
	}

	return PC58_ERR_NONE;
}


/*-------------------------------------------------------------------------
**  PC58_gets() - reads a string from the PC-58 into dest buf using shared 
**     mem interface. If dest buf is too small, data will be lost.
**-------------------------------------------------------------------------
*/
int PC58_gets( int minor, char *s, int s_len )
{
   int inx, len;

   struct pc58_dram_t *sm = (struct pc58_dram_t *) pc58_table[minor].shared_mem;

   len = 0; 
   inx = sm->in_read_inx;  /* get index of next char to read */

   while( 1 )
	{
		if( inx == sm->in_inx_end )   /* queue is empty, return */
		{
			put_user( 0, (s+len));     /* terminate string */
			return PC58_ERR_NONE;
		}

		/* read character into buffer (if we can), and increment index */
		if( len < (s_len-1) )
		{
			put_user( sm->inbuf[inx], (s+len));
			len++;
		}

		if( ++inx >= PC58_BUF_SIZE) /* increment index to next location */
			inx = 0;

		sm->in_read_inx = inx;  /* keep shared memory up-to-date */
   }
}

/*-------------------------------------------------------------------------
**  PC58_clear_inbuf() - Clears the input buffer by 'reading' all the char.
**-------------------------------------------------------------------------
*/
int PC58_clear_inbuf( int minor )
{
   int inx, len;

   struct pc58_dram_t *sm = (struct pc58_dram_t *) pc58_table[minor].shared_mem;

   len     = 0; 
   inx = sm->in_read_inx;  /* get index of next char to read */

   while( 1 )
	{
		if( inx == sm->in_inx_end )   /* queue is empty, return */
			return PC58_ERR_NONE;

		/* limit the loop to size of buffer  */
		if( len++ >= PC58_BUF_SIZE )
			return PC58_ERR_NONE;

		if( ++inx >= PC58_BUF_SIZE) /* increment index to next location */
			inx = 0;

		sm->in_read_inx = inx;  /* keep shared memory up-to-date */
   }
}

/*-------------------------------------------------------------------------
**  PC58_command() - user  _puts() & _gets to write/read a command/reply 
**     to the pc58. 
**-------------------------------------------------------------------------
*/
int PC58_command( int minor, char *command, char *reply,  int r_len )
{
   int rc;

	/* clear input  buffer */
   PC58_clear_inbuf( minor );

	/* write command */
   if( (rc = PC58_puts( minor, command )) != PC58_ERR_NONE)
	   return rc;

	/* sleep 0.1 second */
   PC58csleep( 10 );

	/* reads reply */
   rc = PC58_gets( minor, reply, r_len );
	return rc;
}

/*-------------------------------------------------------------------------
**  PC58_copy_axis_info() - copies the axis infor to the user process.
**-------------------------------------------------------------------------
*/
int PC58_read_axis_info( 
   int minor,                      /* minor device number */
	int axis_inx,                   /* axis index (0 to 7) */
	struct pc58_dram_axis_t *dest   /* destination in user space memory */
)
{
   int cnt;
	struct pc58_dram_t *sm = (struct pc58_dram_t *) pc58_table[minor].shared_mem;

	if( (axis_inx < 0) || (axis_inx > 7))
	   return PC58_ERR_INV_RNG;

if( DEBUG )
	printk( "read_axis_info() addr=0x%lx size=%d\n",  
	   (unsigned long) &sm->motor_status[axis_inx].axis, sizeof(struct pc58_dram_axis_t));
   
	/*------------------------------------------------------------
	** set data_area_update bit & wait until pc58 write new status 
	*/
	outb_p( PC58_CNTL_REG_data_area_update_request, pc58_table[minor].base + PC58_CNTL_REG);
	cnt   = 0;

	while ( inb_p(pc58_table[minor].base + PC58_CNTL_REG) & 
		     PC58_CNTL_REG_data_area_update_request)
	{
		PC58csleep( 1 );
	   if( ++cnt > 5 )
		{
	      printk( "read_axis_info() Time out error!\n"); 
		   return PC58_ERR_TIMEOUT;
		}
	}

	memcpy_tofs( dest, &sm->motor_status[axis_inx].axis, sizeof(struct pc58_dram_axis_t));

	return PC58_ERR_NONE;
}


