/***********************************************************************
**  Device driver private header file for PC-58 -
** --------------------------------------------------------------------
**  Written by:
**     Tony Denault
**     Institute for Astronomy, University of Hawaii
**     2680 Woodlawn Drive, Honolulu, HI 96822
**     Ph: (808)956-8101
**     email: denault@irtf.ifa.hawaii.edu
************************************************************************
*/

#ifndef _LINUX_PC58_H
#define _LINUX_PC58_H

/*------------------------------------------------------------
** pc58parms_t - a general structure used to pass arguments
**               & return values to the ioctl() call.
**------------------------------------------------------------
*/
struct pc58parms_t
{
   int ret;             /* return code    */
   unsigned long l[4];  /* array of longs */
   void *        p[2];  /* pointers */
};

/*------------------------------------------------------------
** pc58_table_t - Each minor device configuration is keep in
**                this table structure.
**------------------------------------------------------------
*/
struct pc58_table_t {
   int base;                     /* base I/O address of board */
	int shared_mem;

	int status_reg;               /* mirror contents of status register */
	int slip_reg;                 /* contents of slip register */
	int done_reg;                 /* contents of done flag  register */
	int limit_reg;                /* contents of limit flag  register */
	int home_reg;                 /* contents of home switch register */
	
	int dio_lo;                   /* mirrors io bits 0-7  */
	int dio_hi;                   /* mirrors io bits 8-15 */
};

/*------------------------------------------------------------
** pc58_dram_t - how the Duel ported ram is organized 
**------------------------------------------------------------
*/
#define PC58_BUF_SIZE     256

struct pc58_dram_axis_t {
	long  encoder_pos;
	long  command_pos;
	long  command_vel;
	long  accel;
	long  max_vel;
	long  base_vel;
	float filter_gain;
	long  filter_pole;
	long  filter_zero;
};

union pc58_axis_u 
{ 
	char status_block[128];
	struct pc58_dram_axis_t axis;
};

struct pc58_dram_t {

                                /* READ Only Shared Memory */
   volatile short in_inx_end;   /* End of  inbuf[]. Where pc58 writes new char */
   volatile short out_inx_end;  /* End of outbuf[]. Where pc58 reads next char */
   volatile short inbuf[PC58_BUF_SIZE];  /* Input Buffer */

	short unused_1[254];

	volatile union pc58_axis_u motor_status[8];

                                   /* READ/WRITE Shared Memory */
   volatile short out_write_inx;   /* Index to outbuf[]. where host writes new char */
   volatile short in_read_inx;     /* Index to  inbuf[]. where host reads new char */
   volatile short outbuf[PC58_BUF_SIZE]; /* Output Buffer */
};


/*----------------------------------
** ERR_CODES for PC58 Driver
**----------------------------------
*/
#define PC58_ERR_NONE                      0
#define PC58_ERR_IOCTL                    -1
#define PC58_ERR_FILE_OPEN                -2
#define PC58_ERR_FILE_READ                -3
#define PC58_ERR_BAD_FILE                 -4
#define PC58_ERR_TIMEOUT                  -5
#define PC58_ERR_INV_FUN                  -6
#define PC58_ERR_INV_RNG                  -7
#define PC58_ERR_INTERNAL                 -8

/*----------------------------------
** defines for ioctrl() commands
**----------------------------------
*/
#define PC58Test               (1)
#define PC58GetConfig          (2)
#define PC58MutTest            (3)
#define PC58Inb                (4)
#define PC58Outb               (5)
#define PC58GetSMValues        (6)
#define PC58Puts               (7)
#define PC58Gets               (8)
#define PC58ReadAxisInfo       (9)
#define PC58Command           (10)
#define PC58ClearSlipFlag     (11)
#define PC58ClearDoneFlag     (12)
#define PC58ClearLimitFlag    (13)
#define PC58ClearHomeFlag     (14)

/*----------------------------------
** defines for IO register interface
**----------------------------------
*/
#define PC58_NUM_IOPORTS         8

/* IO Registers on pc58
**-----------------------
*/

#define PC58_CNTL_REG            0  /* Control Register - Read/Write */

#define PC58_STATUS_REG          1  /* Status Register     - Read only */
#define PC58_DIO_LO_REG          2  /* User DIO bit 0-7    - Read only */
#define PC58_SLIP_REG            3  /* Slip Flags          - Read only */
#define PC58_DONE_REG            4  /* Done Flags          - Read only */
#define PC58_DIO_HI_REG          5  /* User DIO bit 8-13   - Read only */
#define PC58_LIMIT_REG           6  /* Limit Switch Status - Read only */
#define PC58_HOME_REG            7  /* Home Switch Status  - Read only */

#define PC58_DPRAM_LO_REG        6  /* Dual Port RAM Addr (A15-A12) - Write only */
#define PC58_DPRAM_HI_REG        7  /* Dual Port RAM Addr (A23-A16) - Write only */

/*----------------------------
** Bits for control register
*/

#define PC58_CNTL_REG_data_area_update_request   0x01    
#define PC58_CNTL_REG_unused_1                   0x02    
#define PC58_CNTL_REG_encoder_slip_int_enb       0x04    
#define PC58_CNTL_REG_limit_register_int_enb     0x08    
#define PC58_CNTL_REG_done_reg_int_enb           0x10
#define PC58_CNTL_REG_int_request                0x20
#define PC58_CNTL_REG_unused_2                   0x40
#define PC58_CNTL_REG_int_request_enb            0x80

/*--------------------------
** Bits for status register
*/

#define PC58_STATUS_REG_command_error            0x01    
#define PC58_STATUS_REG_initialized              0x02    
#define PC58_STATUS_REG_encoder_slip             0x04    
#define PC58_STATUS_REG_overtravel               0x08    
#define PC58_STATUS_REG_done                     0x10
#define PC58_STATUS_REG_int_req_status_to        0x20
#define PC58_STATUS_REG_dir_int_req_from         0x40
#define PC58_STATUS_REG_int_req_status           0x80

/*-----------------------------------------
** Slip, Done, limit, home registers flags 
*/

#define PC58_axis_X                              0x01    
#define PC58_axis_Y                              0x02    
#define PC58_axis_Z                              0x04    
#define PC58_axis_T                              0x08    
#define PC58_axis_U                              0x10    
#define PC58_axis_V                              0x20    
#define PC58_axis_R                              0x40    
#define PC58_axis_S                              0x80    

/*----------------------------------
** define for iocntl commands 
**----------------------------------
*/
#define PC58TEST                (1)
#define PC58GETCONFIG           (2)
#define PC58MUTTEST             (3)
#define PC58INB                 (4)
#define PC58OUTB                (5)

/*----------------------------------
** prototype for pc58nice.h
**----------------------------------
*/
int  pc58_inb_p( int fd, unsigned short port );
void pc58_outb_p( int fd, char value, unsigned short port );
int  pc58_getConfig( int fd, struct pc58_table_t *config );
int  pc58_mutTest( int fd, int csec, int id );

int pc58_puts( int fd, char *s );
int pc58_gets( int fd, char *s, int s_size );
int pc58_command( int fd, char *command, char *reply, int r_size );
int pc58_readAxisInfo( int fd, int inx, struct pc58_dram_axis_t * dest_buf );
int pc58_clearSlipFlag( int fd, int mask);
int pc58_clearDoneFlag( int fd, int mask);
int pc58_clearLimitFlag( int fd, int mask);
int pc58_clearHomeFlag( int fd, int mask);

#endif /*  _LINUX_PC58_H  */

